"""Module containing constants and dataclasses for the `tm_devices` package."""

import re

from dataclasses import dataclass
from types import MappingProxyType
from typing import Final, FrozenSet, List, Mapping, Optional, Tuple

from pyvisa import constants as pyvisa_constants

from tm_devices.helpers.dataclass_mixins import (
    AsDictionaryMixin,
    AsDictionaryUseEnumNameUseCustEnumStrValueMixin,
)
from tm_devices.helpers.enums import ConnectionTypes, DeviceTypes, SupportedModels
from tm_devices.helpers.standalone_functions import validate_address


@dataclass
class USBTMCConfiguration:
    """Dataclass for holding USBTMC configuration information.

    Each instance of the dataclass holds a Vendor ID and a Model ID to allow easier creation of
    USBTMC VISA resource expressions.
    """

    vendor_id: str
    """The hexadecimal Vendor ID used to create the USBTMC resource expression."""
    model_id: str
    """The hexadecimal Model ID used to create the USBTMC resource expression."""


class _ConfigEntryEnvStrMixin(AsDictionaryMixin):
    """Implements __str__ dunder method to format configuration information as an env variable.

    Handles nested classes through a prefix mapping stored in CONFIG_CLASS_STR_PREFIX_MAPPING.
    Changes to the mapping need to be reversible in the dm_config_parser which needs to bundle
    prefixed values under their own dictionary (to mimic the return type from the toml/yaml config
    files -> ``Dict[str, Union[str, int, Dict[str,...]]]``).

    So make sure to also modify DMConfigParser's _CONFIG_NESTED_DICT_MAPPING class property.
    """

    def __str__(self) -> str:
        """Return a single line of comma separated, key-value pairs for an environment variable."""
        ret_list: List[str] = []
        for key, val in self.to_dict(ignore_none=True).items():
            # must pull class type from actual value, not the to_dict()'s representation
            if (
                prefix := CONFIG_CLASS_STR_PREFIX_MAPPING.get(type(getattr(self, key)))
            ) is not None:
                if any(val.values()):
                    # still need to ignore the None values
                    str_pair = ",".join(
                        f"{prefix}{key2}={val2}" for key2, val2 in val.items() if val2 is not None
                    )
                else:
                    continue  # pragma: no cover
            else:
                str_pair = f"{key}={val}"
            ret_list.append(str_pair)
        return ",".join(ret_list)


@dataclass
class SerialConfig(AsDictionaryUseEnumNameUseCustEnumStrValueMixin, _ConfigEntryEnvStrMixin):
    """Serial configuration properties for connecting to a device over SERIAL (ASRL).

    Notes:
        - `baud_rate`: The baud rate controls the communication frequency.
            e.g. 9600
        - `data_bits`: The number of data bits in each character.
            One of [5, 6, 7, 8].
        - `flow_control`: Control for pausing/resuming data stream between slower devices.
            One of ``SerialConfig.FlowControl.[none|xon_xoff|dtr_dsr|rts_cts]``.
        - `parity`: Parity controls checksum bit (added to each data character) behavior.
            One of ``SerialConfig.Parity.[none|odd|even|mark|space]``.
        - `stop_bits`: Number of bits to use to indicate end of a character.
            One of ``SerialConfig.StopBits.[one|one_and_a_half|two]``.
        - `end_input`: Character(s) to indicate the end of a message transmission.
            One of ``SerialConfig.Termination.[termination_break|termination_char|last_bit|none]``.
    """

    FlowControl = pyvisa_constants.ControlFlow
    """A convenience enumeration listing the options for control flow for a serial device.

    It is used to set the
    [`flow_control`][tm_devices.helpers.constants_and_dataclasses.SerialConfig.flow_control]
    attribute when creating a
    [`SerialConfig`][tm_devices.helpers.constants_and_dataclasses.SerialConfig] object.
    """
    Parity = pyvisa_constants.Parity
    """A convenience enumeration listing the options for parity for a serial device.

    It is used to set the
    [`parity`][tm_devices.helpers.constants_and_dataclasses.SerialConfig.parity]
    attribute when creating a
    [`SerialConfig`][tm_devices.helpers.constants_and_dataclasses.SerialConfig] object.
    """
    StopBits = pyvisa_constants.StopBits
    """A convenience enumeration listing the options for stop bits for a serial device.

    It is used to set the
    [`stop_bits`][tm_devices.helpers.constants_and_dataclasses.SerialConfig.stop_bits]
    attribute when creating a
    [`SerialConfig`][tm_devices.helpers.constants_and_dataclasses.SerialConfig] object.
    """
    Termination = pyvisa_constants.SerialTermination
    """A convenience enumeration listing the available methods for terminating a serial transfer.

    It is used to set the
    [`end_input`][tm_devices.helpers.constants_and_dataclasses.SerialConfig.end_input]
    attribute when creating a
    [`SerialConfig`][tm_devices.helpers.constants_and_dataclasses.SerialConfig] object.
    """
    baud_rate: Optional[int] = None
    """The baud rate controls the communication frequency."""
    data_bits: Optional[int] = None
    """The number of data bits in each character.

    One of ``[5, 6, 7, 8]``.
    """
    flow_control: Optional[FlowControl] = None
    """Control for pausing/resuming data stream between slower devices.

    One of ``SerialConfig.FlowControl.[none|xon_xoff|dtr_dsr|rts_cts]``.
    """
    parity: Optional[Parity] = None
    """Parity adds a checksum bit to each data character.

    This checksum bit enables the target device to determine whether the data was received
    correctly. The parity is used with every frame transmitted and received on a serial session.

    One of ``SerialConfig.Parity.[none|odd|even|mark|space]``.
    """
    stop_bits: Optional[StopBits] = None
    """Number of bits to use to indicate end of a character.

    The number of stop bits that indicate the end of a frame on a serial resource.

    One of ``SerialConfig.StopBits.[one|one_and_a_half|two]``.
    """
    end_input: Optional[Termination] = None
    """Character(s) to indicate the end of a message transmission.

    One of ``SerialConfig.Termination.[termination_break|termination_char|last_bit|none]``.
    """

    def __post_init__(self) -> None:
        """Convert from string to expected type and translate enum string names into the values."""
        if isinstance(self.baud_rate, str):
            self.baud_rate = int(self.baud_rate)
        if isinstance(self.data_bits, str):
            self.data_bits = int(self.data_bits)
        if isinstance(self.flow_control, str):
            self.flow_control = getattr(self.FlowControl, self.flow_control)
        if isinstance(self.parity, str):
            self.parity = getattr(self.Parity, self.parity)
        if isinstance(self.stop_bits, str):
            self.stop_bits = getattr(self.StopBits, self.stop_bits)
        if isinstance(self.end_input, str):
            self.end_input = getattr(self.Termination, self.end_input)


@dataclass(frozen=True)
class DeviceConfigEntry(AsDictionaryUseEnumNameUseCustEnumStrValueMixin, _ConfigEntryEnvStrMixin):
    """Dataclass for holding configuration information for a single device."""

    device_type: DeviceTypes
    """The specific device type defined in the config entry."""
    address: str
    """The address used to connect to the device.

    Notes:
        - TCPIP: IP address or the hostname.
        - SOCKET: IP address or the hostname (must define `lan_port` field).
        - REST_API: IP address or the hostname (must define `lan_port` field).
        - SERIAL/ASRL: serial COM port number.
        - USB: use expression format `f"{model}-{serial_number}"` (ex: `"MSO24-ABC0123"`).
    """
    connection_type: ConnectionTypes = ConnectionTypes.TCPIP
    """The specific type of connection defined in the config entry."""
    alias: Optional[str] = None
    """An optional key/name used to retrieve this devices from the DeviceManager."""
    lan_port: Optional[int] = None
    """The port number to connect on, used for SOCKET/REST_API connections."""
    serial_config: Optional[SerialConfig] = None
    """Serial configuration properties for connecting to a device over SERIAL (ASRL)."""
    device_driver: Optional[str] = None
    """The name of the Python driver to use for the device (required for connection_type=REST_API, ignored otherwise)"""  # noqa: E501

    # pylint: disable=too-many-branches
    def __post_init__(self) -> None:  # noqa: PLR0912,C901
        """Validate data after creation.

        Raises:
            TypeError: Indicates a bad enum type.
            ValueError: Indicates a bad input value.
        """
        # manually handle type conversion from parsing from config file
        # (this is the correct way to edit the frozen data)
        if self.lan_port is not None and isinstance(self.lan_port, str):
            object.__setattr__(self, "lan_port", int(self.lan_port))
        if not isinstance(self.address, str):  # pyright: ignore[reportUnnecessaryIsInstance]
            object.__setattr__(self, "address", str(self.address))

        # Validate the connection and interface types
        try:
            # device and connections are inside this try clause to raise a TypeError on bad values.
            # Convert them from strings (parsed from config/env) to appropriate enum value.
            if not isinstance(self.device_type, DeviceTypes):  # pyright: ignore[reportUnnecessaryIsInstance]
                object.__setattr__(self, "device_type", DeviceTypes(self.device_type))
            if not isinstance(self.connection_type, ConnectionTypes):  # pyright: ignore[reportUnnecessaryIsInstance]
                object.__setattr__(self, "connection_type", ConnectionTypes(self.connection_type))

            # While a SerialConfig is not frozen, if not created here then it cannot be added later.
            # A serial_config must be created if the connection_type is SERIAL (ASRL).
            if self.connection_type == ConnectionTypes.SERIAL and self.serial_config is None:
                object.__setattr__(self, "serial_config", SerialConfig())

            if self.device_type not in VALID_DEVICE_CONNECTION_TYPES:
                msg = (
                    f"{self.device_type.name} is not a valid device type. Valid device types: "
                    f"{VALID_DEVICE_CONNECTION_TYPES}"
                )
                raise TypeError(msg)
            if self.connection_type not in VALID_DEVICE_CONNECTION_TYPES[self.device_type]:
                msg = (
                    f"{self.connection_type.name} is not a valid "
                    f"{self.device_type} connection type. Valid connection types: "
                    f"{[x.name for x in VALID_DEVICE_CONNECTION_TYPES[self.device_type]]}"
                )
                raise TypeError(msg)
        except ValueError as error:
            # this is from an invalid enum name
            raise TypeError(*error.args)  # noqa: TRY200,B904

        # Check if a specific device driver is needed to be specified
        if self.connection_type == ConnectionTypes.REST_API and self.device_driver is None:
            msg = (
                f'Device configuration entry "{self}" specified a '
                f"REST_API connection without providing the Python device driver name to use."
            )
            raise ValueError(msg)

        if self.connection_type == ConnectionTypes.SOCKET and self.lan_port is None:
            msg = (
                f'Device configuration entry "{self}" specified a '
                f"socket connection without a port number."
            )
            raise ValueError(msg)

        # validate address
        try:
            if self.connection_type in [
                ConnectionTypes.TCPIP,
                ConnectionTypes.SOCKET,
                ConnectionTypes.REST_API,
            ]:
                validate_address(self.address)
            elif (  # pylint: disable=too-many-boolean-expressions
                (self.connection_type == ConnectionTypes.USB and "-" not in self.address)
                or (self.connection_type == ConnectionTypes.SERIAL and not self.address.isdigit())
                or (
                    self.connection_type == ConnectionTypes.GPIB
                    and not (self.address.isdigit() and 1 <= int(self.address) <= MAX_GPIB_ADDRESS)
                )
            ):
                raise ValueError
        except ValueError as exc:
            msg = f"""Found invalid value of `address={self.address}`
        Must provide an valid address format for the `connection_type={self.connection_type.value}':
            TCPIP: IP address or the hostname.
            SOCKET: IP address or the hostname (must define `lan_port` field).
            REST_API: IP address or the hostname (must define `lan_port` field).
            SERIAL/ASRL: serial COM port number.
            GPIB: address number (1 <= n <= 30).
            USB: use expression format "<model>-<serial_number>" (ex: "MSO24-ABC0123").
                """
            raise ValueError(msg) from exc

        # Validate serial config
        if self.serial_config is not None:
            # make sure serial attributes with integer input are in valid range
            if (
                self.serial_config.baud_rate is not None
                and self.serial_config.baud_rate not in VALID_SERIAL_BAUD
            ):
                msg = (
                    f"Unexpected baud rate of {self.serial_config.baud_rate}, "
                    f"must be in {sorted(VALID_SERIAL_BAUD)}"
                )
                raise ValueError(msg)
            if (
                self.serial_config.data_bits is not None
                and self.serial_config.data_bits not in VALID_SERIAL_DATA_BITS
            ):
                msg = (
                    f"Unexpected data bits of {self.serial_config.data_bits}, "
                    f"must be in {sorted(VALID_SERIAL_DATA_BITS)}"
                )
                raise ValueError(msg)

    def get_address_expression(self) -> str:
        """Return an address expression.

        This is not the resource expression, it is used to help catch duplicate addresses.
        """
        return (
            self.connection_type.value + " " + self.address + ""
            if self.lan_port is None
            else f":{self.lan_port}"
        )

    def get_visa_resource_expression(self) -> str:
        """Construct the VISA resource expression for the device.

        Example visa expressions:

            - TCPIP0::MSO58-3000260::instr0::INSTR
            - USB0::0x0699::0x0522::3000260::INSTR
            - TCPIP0::192.168.0.1::4000::SOCKET
            - ASRL3::INSTR  (serial on COM port 3)

        Returns:
            A string containing the VISA resource expression.

        Raises:
            NotImplementedError: Indicates that the USBTMC lookup dictionary needs to be
                updated to support the current device.
            ValueError: Indicates that the connection type specified in the config is not valid.
        """
        # If USBTMC for the address check if connected
        if self.connection_type == ConnectionTypes.USB:
            full_model, serial = self.address.split("-", 1)
            # local import here to avoid circular imports
            # pylint: disable=import-outside-toplevel,cyclic-import
            from tm_devices.helpers.functions import get_model_series

            model_series = get_model_series(full_model)
            try:
                usbtmc_config = USB_MODEL_ID_LOOKUP[model_series]

                # This commented out block of code can help deal with devices where the specific
                # model number is the USBTMC model_id but the Python driver is shared between
                # multiple model numbers. It is currently not needed, if this is needed in the
                # future this code block can be uncommented.
                # RELIC # if usbtmc_config.model_id == _USBTMC_UNKNOWN_MODEL:
                # RELIC #     usbtmc_config = USBTMCConfiguration(
                # RELIC #         vendor_id=usbtmc_config.vendor_id,
                # RELIC #         model_id="0x" + (full_model.split(" ", maxsplit=1)[-1])[:4],
                # RELIC #     )

                resource_expr = (
                    f"USB0::{usbtmc_config.vendor_id}::{usbtmc_config.model_id}::{serial}::0::INSTR"
                )
            except KeyError as exc:
                msg = (
                    f"USBTMC model ID has not been configured for the "
                    f"{model_series} model series\nconfig = {self}"
                )
                raise NotImplementedError(msg) from exc
        elif self.connection_type == ConnectionTypes.SOCKET:
            resource_expr = f"TCPIP0::{self.address}::{self.lan_port}::SOCKET"
        elif self.connection_type == ConnectionTypes.TCPIP:
            resource_expr = f"TCPIP0::{self.address}::inst0::INSTR"
        elif self.connection_type == ConnectionTypes.GPIB:
            resource_expr = f"GPIB0::{self.address}::INSTR"
        elif self.connection_type == ConnectionTypes.SERIAL:
            resource_expr = f"ASRL{self.address}::INSTR"
        elif self.connection_type == ConnectionTypes.MOCK:  # pragma: no cover
            resource_expr = f"MOCK0::{self.address}::INSTR"
        else:
            msg = (
                f"{self.connection_type.value} is not a valid connection "
                f"type for a VISA resource."
            )
            raise ValueError(msg)
        return resource_expr


@dataclass
class DMConfigOptions(AsDictionaryMixin):
    """Device Management Configuration options."""

    standalone: Optional[bool] = None
    """An option indicating to use PyVISA-py's pure Python VISA backend."""
    setup_cleanup: Optional[bool] = None
    """An option indicating to run a device's ``cleanup()`` method when opening the connection."""
    teardown_cleanup: Optional[bool] = None
    """An option indicating to run a device's ``cleanup()`` method when closing the connection."""
    verbose_mode: Optional[bool] = None
    """A verbosity flag to turn on more printouts to stdout."""
    verbose_visa: Optional[bool] = None
    """A verbosity flag to enable extremely verbose VISA logging to stdout."""
    retry_visa_connection: Optional[bool] = None
    """A flag to enable retrying the first VISA connection attempt."""
    check_for_updates: Optional[bool] = None
    """A flag indicating if a check for updates for the package should be performed on creation of the DeviceManager."""  # noqa: E501

    def __str__(self) -> str:
        """Complete config entry line for an environment variable."""
        return ",".join(
            [
                opt_key.upper()
                for opt_key, opt_val in self.to_dict(ignore_none=True).items()
                if opt_val
            ]
        )


###############################################################################
#                                CONSTANTS
###############################################################################
# don't include this in the __init__
MAX_GPIB_ADDRESS: Final[int] = 30
# don't include this in the __init__
CONFIG_CLASS_STR_PREFIX_MAPPING: Final = MappingProxyType({SerialConfig: "serial_"})
# don't include this in the __init__
VALID_SERIAL_BAUD: Final[FrozenSet[int]] = frozenset(
    [
        300,
        600,
        1200,
        2400,
        4800,
        9600,
        14400,
        19200,
        28800,
        31250,
        38400,
        57600,
        115200,
    ]
)
# don't include this in the __init__
VALID_SERIAL_DATA_BITS: Final[FrozenSet[int]] = frozenset([5, 6, 7, 8])

PYVISA_PY_BACKEND: Final[str] = "@py"
"""Constant string which indicates to use the pure Python PyVISA-py backend when creating VISA connections."""  # noqa: E501

SYSTEM_DEFAULT_VISA_BACKEND: Final[str] = ""
"""Constant string which indicates to use the current system's default VISA backend when creating VISA connections."""  # noqa: E501

PACKAGE_NAME: Final[str] = "tm_devices"
"""Constant string with the name of this package."""

VISA_RESOURCE_EXPRESSION_REGEX: "Final[re.Pattern[str]]" = re.compile(  # pylint: disable=unsubscriptable-object,useless-suppression
    r"^(\w+)(?:::0X\w+)?::([-.\w]+)(?:::(\w+))?(?:::INST0?)?::(INSTR?|SOCKET)$"
)
"""A regex pattern used to capture pieces of VISA resource expressions."""

UNIT_TEST_TIMEOUT: Final[int] = 50
"""The VISA timeout value to use during unit tests, in milliseconds."""

VALID_DEVICE_CONNECTION_TYPES: Final[Mapping[DeviceTypes, Tuple[ConnectionTypes, ...]]] = (
    MappingProxyType(
        {
            DeviceTypes.AFG: (
                ConnectionTypes.TCPIP,
                ConnectionTypes.USB,
                ConnectionTypes.SOCKET,
                ConnectionTypes.SERIAL,
                ConnectionTypes.GPIB,
            ),
            DeviceTypes.AWG: (
                ConnectionTypes.TCPIP,
                ConnectionTypes.USB,
                ConnectionTypes.SOCKET,
                ConnectionTypes.SERIAL,
                ConnectionTypes.GPIB,
            ),
            DeviceTypes.DAQ: (
                ConnectionTypes.TCPIP,
                ConnectionTypes.USB,
                ConnectionTypes.GPIB,
                ConnectionTypes.SERIAL,
            ),
            DeviceTypes.DMM: (
                ConnectionTypes.TCPIP,
                ConnectionTypes.USB,
                ConnectionTypes.SOCKET,
                ConnectionTypes.SERIAL,
                ConnectionTypes.GPIB,
            ),
            DeviceTypes.MT: (ConnectionTypes.REST_API,),
            DeviceTypes.PSU: (
                ConnectionTypes.TCPIP,
                ConnectionTypes.USB,
                ConnectionTypes.SOCKET,
                ConnectionTypes.SERIAL,
                ConnectionTypes.GPIB,
            ),
            DeviceTypes.SCOPE: (
                ConnectionTypes.TCPIP,
                ConnectionTypes.USB,
                ConnectionTypes.SOCKET,
                ConnectionTypes.SERIAL,
                ConnectionTypes.GPIB,
                ConnectionTypes.MOCK,
            ),
            DeviceTypes.SMU: (
                ConnectionTypes.TCPIP,
                ConnectionTypes.USB,
                ConnectionTypes.SOCKET,
                ConnectionTypes.SERIAL,
                ConnectionTypes.GPIB,
            ),
            DeviceTypes.SS: (
                ConnectionTypes.TCPIP,
                ConnectionTypes.USB,
                ConnectionTypes.SOCKET,
                ConnectionTypes.GPIB,
            ),
        }
    )
)
"""Mapping of each device type to its supported connection types.

Any additions to this class need to be added to the [`tm_devices.helpers.enums.DeviceTypes`][] enum
as well.
"""

# USBTMC configuration defines
_TEKTRONIX_USBTMC_VENDOR_ID: Final[str] = "0x0699"
_KEITHLEY_USBTMC_VENDOR_ID: Final[str] = "0x05E6"
USB_MODEL_ID_LOOKUP: Final[Mapping[str, USBTMCConfiguration]] = MappingProxyType(
    {
        SupportedModels.MDO3K.value: USBTMCConfiguration(
            vendor_id=_TEKTRONIX_USBTMC_VENDOR_ID, model_id="0x0408"
        ),
        SupportedModels.MSO2.value: USBTMCConfiguration(
            vendor_id=_TEKTRONIX_USBTMC_VENDOR_ID, model_id="0x0105"
        ),
        SupportedModels.MSO4.value: USBTMCConfiguration(
            vendor_id=_TEKTRONIX_USBTMC_VENDOR_ID, model_id="0x0527"
        ),
        SupportedModels.MSO4B.value: USBTMCConfiguration(
            vendor_id=_TEKTRONIX_USBTMC_VENDOR_ID, model_id="0x0527"
        ),
        SupportedModels.MSO5.value: USBTMCConfiguration(
            vendor_id=_TEKTRONIX_USBTMC_VENDOR_ID, model_id="0x0522"
        ),
        SupportedModels.MSO5B.value: USBTMCConfiguration(
            vendor_id=_TEKTRONIX_USBTMC_VENDOR_ID, model_id="0x0522"
        ),
        SupportedModels.MSO5LP.value: USBTMCConfiguration(
            vendor_id=_TEKTRONIX_USBTMC_VENDOR_ID, model_id="0x0529"
        ),
        SupportedModels.MSO6.value: USBTMCConfiguration(
            vendor_id=_TEKTRONIX_USBTMC_VENDOR_ID, model_id="0x0528"
        ),
        SupportedModels.MSO6B.value: USBTMCConfiguration(
            vendor_id=_TEKTRONIX_USBTMC_VENDOR_ID, model_id="0x0530"
        ),
        SupportedModels.LPD6.value: USBTMCConfiguration(
            vendor_id=_TEKTRONIX_USBTMC_VENDOR_ID, model_id="0x052F"
        ),
        SupportedModels.AFG3K.value: USBTMCConfiguration(
            vendor_id=_TEKTRONIX_USBTMC_VENDOR_ID, model_id="0x0345"
        ),
        SupportedModels.SMU2450.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2450"
        ),
        SupportedModels.SMU2460.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2460"
        ),
        SupportedModels.SMU2461.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2461"
        ),
        SupportedModels.SMU2470.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2470"
        ),
        SupportedModels.SMU2601A.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2601"
        ),
        SupportedModels.SMU2602A.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2602"
        ),
        SupportedModels.SMU2604A.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2604"
        ),
        SupportedModels.SMU2611A.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2611"
        ),
        SupportedModels.SMU2612A.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2612"
        ),
        SupportedModels.SMU2614A.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2614"
        ),
        SupportedModels.SMU2634A.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2634"
        ),
        SupportedModels.SMU2635A.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2635"
        ),
        SupportedModels.SMU2636A.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2636"
        ),
        SupportedModels.SMU2601B.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2601"
        ),
        SupportedModels.SMU2602B.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2602"
        ),
        SupportedModels.SMU2604B.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2604"
        ),
        SupportedModels.SMU2606B.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2606"
        ),
        SupportedModels.SMU2611B.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2611"
        ),
        SupportedModels.SMU2612B.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2612"
        ),
        SupportedModels.SMU2614B.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2614"
        ),
        SupportedModels.SMU2634B.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2634"
        ),
        SupportedModels.SMU2635B.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2635"
        ),
        SupportedModels.SMU2636B.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2636"
        ),
        SupportedModels.PSU2200.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2200"
        ),
        SupportedModels.PSU2220.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2220"
        ),
        SupportedModels.PSU2230.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2230"
        ),
        SupportedModels.PSU2231.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2231"
        ),
        SupportedModels.PSU2231A.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2231"
        ),
        SupportedModels.PSU2280.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2280"
        ),
        SupportedModels.PSU2281.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x2281"
        ),
        SupportedModels.AWG5200.value: USBTMCConfiguration(
            vendor_id=_TEKTRONIX_USBTMC_VENDOR_ID, model_id="0x0503"
        ),
        SupportedModels.AWG70KA.value: USBTMCConfiguration(
            vendor_id=_TEKTRONIX_USBTMC_VENDOR_ID, model_id="0x0503"
        ),
        SupportedModels.AWG70KB.value: USBTMCConfiguration(
            vendor_id=_TEKTRONIX_USBTMC_VENDOR_ID, model_id="0x0503"
        ),
        SupportedModels.SS3706A.value: USBTMCConfiguration(
            vendor_id=_KEITHLEY_USBTMC_VENDOR_ID, model_id="0x3706"
        ),
    }
)
"""A mapping of model USBTMC info."""



"""The autoset commands module.

These commands are used in the following models:
DPO2K, DPO2KB, MSO2K, MSO2KB

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Commands and Queries:
    ```
    - AUTOSet {EXECute|UNDo}
    - AUTOSet:ENAble {OFF|ON}
    - AUTOSet:ENAble?
    ```
"""

from typing import Optional, TYPE_CHECKING

from ..helpers import SCPICmdRead, SCPICmdWrite

if TYPE_CHECKING:
    from tm_devices.drivers.pi.pi_device import PIDevice


class AutosetEnable(SCPICmdWrite, SCPICmdRead):
    """The ``AUTOSet:ENAble`` command.

    Description:
        - Enables or disables the autoset.

    Usage:
        - Using the ``.query()`` method will send the ``AUTOSet:ENAble?`` query.
        - Using the ``.verify(value)`` method will send the ``AUTOSet:ENAble?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``AUTOSet:ENAble value`` command.

    SCPI Syntax:
        ```
        - AUTOSet:ENAble {OFF|ON}
        - AUTOSet:ENAble?
        ```

    Info:
        - ``ON`` enables the autoset feature.
        - ``OFF`` disables the autoset feature.
    """


class Autoset(SCPICmdWrite, SCPICmdRead):
    """The ``AUTOSet`` command.

    Description:
        - Sets the vertical, horizontal, and trigger controls of the oscilloscope to automatically
          acquire and display the selected waveform.

    Usage:
        - Using the ``.write(value)`` method will send the ``AUTOSet value`` command.

    SCPI Syntax:
        ```
        - AUTOSet {EXECute|UNDo}
        ```

    Info:
        - ``EXECute`` autosets the displayed waveform.
        - ``UNDo`` restores the oscilloscope settings to those present prior to the autoset
          execution.

    Properties:
        - ``.enable``: The ``AUTOSet:ENAble`` command.
    """

    def __init__(self, device: Optional["PIDevice"] = None, cmd_syntax: str = "AUTOSet") -> None:
        super().__init__(device, cmd_syntax)
        self._enable = AutosetEnable(device, f"{self._cmd_syntax}:ENAble")

    @property
    def enable(self) -> AutosetEnable:
        """Return the ``AUTOSet:ENAble`` command.

        Description:
            - Enables or disables the autoset.

        Usage:
            - Using the ``.query()`` method will send the ``AUTOSet:ENAble?`` query.
            - Using the ``.verify(value)`` method will send the ``AUTOSet:ENAble?`` query and raise
              an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``AUTOSet:ENAble value`` command.

        SCPI Syntax:
            ```
            - AUTOSet:ENAble {OFF|ON}
            - AUTOSet:ENAble?
            ```

        Info:
            - ``ON`` enables the autoset feature.
            - ``OFF`` disables the autoset feature.
        """
        return self._enable



"""The display commands module.

These commands are used in the following models:
AFG3K, AFG3KB, AFG3KC

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Commands and Queries:
    ```
    - DISPlay:BRIGHtness {|MINimum|MAXimum}?[MINimum|MAXimum]
    - DISPlay:CONTrast {<contrast>|MINimum|MAXimum}
    - DISPlay:CONTrast?
    - DISPlay:SAVer:IMMediate
    - DISPlay:SAVer:STATe {ON|OFF|<NR1>}
    - DISPlay:SAVer:STATe?
    - DISPlay:WINDow:TEXT:CLEar
    - DISPlay:WINDow:TEXT:DATA <string>
    - DISPlay:WINDow:TEXT:DATA?
    ```
"""

from typing import Optional, TYPE_CHECKING

from ..helpers import SCPICmdRead, SCPICmdWrite, SCPICmdWriteNoArguments

if TYPE_CHECKING:
    from tm_devices.drivers.pi.pi_device import PIDevice


class DisplayWindowTextData(SCPICmdWrite, SCPICmdRead):
    """The ``DISPlay:WINDow:TEXT:DATA`` command.

    Description:
        - The DISPlay[``:WINDow``]``:TEXT``[``:DATA``] command displays a text message on the
          instrument screen. The DISPlay[``:WINDow``]``:TEXT``[``:DATA``]? query returns the text
          string currently displayed on the instrument screen. The displayable characters are ASCII
          codes 32 through 126, and the instrument can display approximately 64 characters.

    Usage:
        - Using the ``.query()`` method will send the ``DISPlay:WINDow:TEXT:DATA?`` query.
        - Using the ``.verify(value)`` method will send the ``DISPlay:WINDow:TEXT:DATA?`` query and
          raise an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``DISPlay:WINDow:TEXT:DATA value``
          command.

    SCPI Syntax:
        ```
        - DISPlay:WINDow:TEXT:DATA <string>
        - DISPlay:WINDow:TEXT:DATA?
        ```
    """


class DisplayWindowTextClear(SCPICmdWriteNoArguments):
    """The ``DISPlay:WINDow:TEXT:CLEar`` command.

    Description:
        - This command clears the text message from the display screen.

    Usage:
        - Using the ``.write()`` method will send the ``DISPlay:WINDow:TEXT:CLEar`` command.

    SCPI Syntax:
        ```
        - DISPlay:WINDow:TEXT:CLEar
        ```
    """


class DisplayWindowText(SCPICmdRead):
    """The ``DISPlay:WINDow:TEXT`` command tree.

    Usage:
        - Using the ``.query()`` method will send the ``DISPlay:WINDow:TEXT?`` query.
        - Using the ``.verify(value)`` method will send the ``DISPlay:WINDow:TEXT?`` query and raise
          an AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.clear``: The ``DISPlay:WINDow:TEXT:CLEar`` command.
        - ``.data``: The ``DISPlay:WINDow:TEXT:DATA`` command.
    """

    def __init__(self, device: Optional["PIDevice"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        self._clear = DisplayWindowTextClear(device, f"{self._cmd_syntax}:CLEar")
        self._data = DisplayWindowTextData(device, f"{self._cmd_syntax}:DATA")

    @property
    def clear(self) -> DisplayWindowTextClear:
        """Return the ``DISPlay:WINDow:TEXT:CLEar`` command.

        Description:
            - This command clears the text message from the display screen.

        Usage:
            - Using the ``.write()`` method will send the ``DISPlay:WINDow:TEXT:CLEar`` command.

        SCPI Syntax:
            ```
            - DISPlay:WINDow:TEXT:CLEar
            ```
        """
        return self._clear

    @property
    def data(self) -> DisplayWindowTextData:
        """Return the ``DISPlay:WINDow:TEXT:DATA`` command.

        Description:
            - The DISPlay[``:WINDow``]``:TEXT``[``:DATA``] command displays a text message on the
              instrument screen. The DISPlay[``:WINDow``]``:TEXT``[``:DATA``]? query returns the
              text string currently displayed on the instrument screen. The displayable characters
              are ASCII codes 32 through 126, and the instrument can display approximately 64
              characters.

        Usage:
            - Using the ``.query()`` method will send the ``DISPlay:WINDow:TEXT:DATA?`` query.
            - Using the ``.verify(value)`` method will send the ``DISPlay:WINDow:TEXT:DATA?`` query
              and raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``DISPlay:WINDow:TEXT:DATA value``
              command.

        SCPI Syntax:
            ```
            - DISPlay:WINDow:TEXT:DATA <string>
            - DISPlay:WINDow:TEXT:DATA?
            ```
        """
        return self._data


class DisplayWindow(SCPICmdRead):
    """The ``DISPlay:WINDow`` command tree.

    Usage:
        - Using the ``.query()`` method will send the ``DISPlay:WINDow?`` query.
        - Using the ``.verify(value)`` method will send the ``DISPlay:WINDow?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.text``: The ``DISPlay:WINDow:TEXT`` command tree.
    """

    def __init__(self, device: Optional["PIDevice"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        self._text = DisplayWindowText(device, f"{self._cmd_syntax}:TEXT")

    @property
    def text(self) -> DisplayWindowText:
        """Return the ``DISPlay:WINDow:TEXT`` command tree.

        Usage:
            - Using the ``.query()`` method will send the ``DISPlay:WINDow:TEXT?`` query.
            - Using the ``.verify(value)`` method will send the ``DISPlay:WINDow:TEXT?`` query and
              raise an AssertionError if the returned value does not match ``value``.

        Sub-properties:
            - ``.clear``: The ``DISPlay:WINDow:TEXT:CLEar`` command.
            - ``.data``: The ``DISPlay:WINDow:TEXT:DATA`` command.
        """
        return self._text


class DisplaySaverState(SCPICmdWrite, SCPICmdRead):
    """The ``DISPlay:SAVer:STATe`` command.

    Description:
        - This command sets or queries the screen saver setting of the LCD display. When enabled,
          the screen saver function starts automatically if no operations are applied to the
          instrument front panel for five minutes.

    Usage:
        - Using the ``.query()`` method will send the ``DISPlay:SAVer:STATe?`` query.
        - Using the ``.verify(value)`` method will send the ``DISPlay:SAVer:STATe?`` query and raise
          an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``DISPlay:SAVer:STATe value`` command.

    SCPI Syntax:
        ```
        - DISPlay:SAVer:STATe {ON|OFF|<NR1>}
        - DISPlay:SAVer:STATe?
        ```
    """


class DisplaySaverImmediate(SCPICmdWriteNoArguments):
    """The ``DISPlay:SAVer:IMMediate`` command.

    Description:
        - This command sets the screen saver state to ON, regardless of the
          ``DISPlay:SAVer``[``:STATe``]? command setting. The screen saver is enabled immediately
          (without waiting for five minutes).

    Usage:
        - Using the ``.write()`` method will send the ``DISPlay:SAVer:IMMediate`` command.

    SCPI Syntax:
        ```
        - DISPlay:SAVer:IMMediate
        ```
    """


class DisplaySaver(SCPICmdRead):
    """The ``DISPlay:SAVer`` command tree.

    Usage:
        - Using the ``.query()`` method will send the ``DISPlay:SAVer?`` query.
        - Using the ``.verify(value)`` method will send the ``DISPlay:SAVer?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.immediate``: The ``DISPlay:SAVer:IMMediate`` command.
        - ``.state``: The ``DISPlay:SAVer:STATe`` command.
    """

    def __init__(self, device: Optional["PIDevice"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        self._immediate = DisplaySaverImmediate(device, f"{self._cmd_syntax}:IMMediate")
        self._state = DisplaySaverState(device, f"{self._cmd_syntax}:STATe")

    @property
    def immediate(self) -> DisplaySaverImmediate:
        """Return the ``DISPlay:SAVer:IMMediate`` command.

        Description:
            - This command sets the screen saver state to ON, regardless of the
              ``DISPlay:SAVer``[``:STATe``]? command setting. The screen saver is enabled
              immediately (without waiting for five minutes).

        Usage:
            - Using the ``.write()`` method will send the ``DISPlay:SAVer:IMMediate`` command.

        SCPI Syntax:
            ```
            - DISPlay:SAVer:IMMediate
            ```
        """
        return self._immediate

    @property
    def state(self) -> DisplaySaverState:
        """Return the ``DISPlay:SAVer:STATe`` command.

        Description:
            - This command sets or queries the screen saver setting of the LCD display. When
              enabled, the screen saver function starts automatically if no operations are applied
              to the instrument front panel for five minutes.

        Usage:
            - Using the ``.query()`` method will send the ``DISPlay:SAVer:STATe?`` query.
            - Using the ``.verify(value)`` method will send the ``DISPlay:SAVer:STATe?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``DISPlay:SAVer:STATe value``
              command.

        SCPI Syntax:
            ```
            - DISPlay:SAVer:STATe {ON|OFF|<NR1>}
            - DISPlay:SAVer:STATe?
            ```
        """
        return self._state


class DisplayContrast(SCPICmdWrite, SCPICmdRead):
    """The ``DISPlay:CONTrast`` command.

    Description:
        - This command sets or queries the contrast of the LCD display.

    Usage:
        - Using the ``.query()`` method will send the ``DISPlay:CONTrast?`` query.
        - Using the ``.verify(value)`` method will send the ``DISPlay:CONTrast?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``DISPlay:CONTrast value`` command.

    SCPI Syntax:
        ```
        - DISPlay:CONTrast {<contrast>|MINimum|MAXimum}
        - DISPlay:CONTrast?
        ```
    """


class DisplayBrightness(SCPICmdWrite):
    """The ``DISPlay:BRIGHtness`` command.

    Description:
        - This command sets or queries the brightness of the LCD display.

    Usage:
        - Using the ``.write(value)`` method will send the ``DISPlay:BRIGHtness value`` command.

    SCPI Syntax:
        ```
        - DISPlay:BRIGHtness {|MINimum|MAXimum}?[MINimum|MAXimum]
        ```
    """


class Display(SCPICmdRead):
    """The ``DISPlay`` command tree.

    Usage:
        - Using the ``.query()`` method will send the ``DISPlay?`` query.
        - Using the ``.verify(value)`` method will send the ``DISPlay?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.brightness``: The ``DISPlay:BRIGHtness`` command.
        - ``.contrast``: The ``DISPlay:CONTrast`` command.
        - ``.saver``: The ``DISPlay:SAVer`` command tree.
        - ``.window``: The ``DISPlay:WINDow`` command tree.
    """

    def __init__(self, device: Optional["PIDevice"] = None, cmd_syntax: str = "DISPlay") -> None:
        super().__init__(device, cmd_syntax)
        self._brightness = DisplayBrightness(device, f"{self._cmd_syntax}:BRIGHtness")
        self._contrast = DisplayContrast(device, f"{self._cmd_syntax}:CONTrast")
        self._saver = DisplaySaver(device, f"{self._cmd_syntax}:SAVer")
        self._window = DisplayWindow(device, f"{self._cmd_syntax}:WINDow")

    @property
    def brightness(self) -> DisplayBrightness:
        """Return the ``DISPlay:BRIGHtness`` command.

        Description:
            - This command sets or queries the brightness of the LCD display.

        Usage:
            - Using the ``.write(value)`` method will send the ``DISPlay:BRIGHtness value`` command.

        SCPI Syntax:
            ```
            - DISPlay:BRIGHtness {|MINimum|MAXimum}?[MINimum|MAXimum]
            ```
        """
        return self._brightness

    @property
    def contrast(self) -> DisplayContrast:
        """Return the ``DISPlay:CONTrast`` command.

        Description:
            - This command sets or queries the contrast of the LCD display.

        Usage:
            - Using the ``.query()`` method will send the ``DISPlay:CONTrast?`` query.
            - Using the ``.verify(value)`` method will send the ``DISPlay:CONTrast?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``DISPlay:CONTrast value`` command.

        SCPI Syntax:
            ```
            - DISPlay:CONTrast {<contrast>|MINimum|MAXimum}
            - DISPlay:CONTrast?
            ```
        """
        return self._contrast

    @property
    def saver(self) -> DisplaySaver:
        """Return the ``DISPlay:SAVer`` command tree.

        Usage:
            - Using the ``.query()`` method will send the ``DISPlay:SAVer?`` query.
            - Using the ``.verify(value)`` method will send the ``DISPlay:SAVer?`` query and raise
              an AssertionError if the returned value does not match ``value``.

        Sub-properties:
            - ``.immediate``: The ``DISPlay:SAVer:IMMediate`` command.
            - ``.state``: The ``DISPlay:SAVer:STATe`` command.
        """
        return self._saver

    @property
    def window(self) -> DisplayWindow:
        """Return the ``DISPlay:WINDow`` command tree.

        Usage:
            - Using the ``.query()`` method will send the ``DISPlay:WINDow?`` query.
            - Using the ``.verify(value)`` method will send the ``DISPlay:WINDow?`` query and raise
              an AssertionError if the returned value does not match ``value``.

        Sub-properties:
            - ``.text``: The ``DISPlay:WINDow:TEXT`` command tree.
        """
        return self._window



"""The saveon commands module.

These commands are used in the following models:
DPO5K, DPO5KB, DPO70KC, DPO70KD, DPO70KDX, DPO70KSX, DPO7K, DPO7KC, DSA70KC, DSA70KD, MSO5K, MSO5KB,
MSO70KC, MSO70KDX

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Commands and Queries:
    ```
    - SAVEON {RESET}
    - SAVEON:COUNt?
    - SAVEON:FILE:AUTOInc {<NR1>|ON|OFF}
    - SAVEON:FILE:AUTOInc?
    - SAVEON:FILE:COUNt <NR3>
    - SAVEON:FILE:COUNt?
    - SAVEON:FILE:DEST <string>
    - SAVEON:FILE:DEST?
    - SAVEON:FILE:NAME <string>
    - SAVEON:FILE:NAME?
    - SAVEON:FILE:TYPE {AUTO|CUSTOM}
    - SAVEON:FILE:TYPE?
    - SAVEON:IMAGe {<NR1>|OFF|ON}
    - SAVEON:IMAGe?
    - SAVEON:LIMit {<NR1>|OFF|ON}
    - SAVEON:LIMit?
    - SAVEON:MASK {<NR1>|OFF|ON}
    - SAVEON:MASK?
    - SAVEON:MEASUrement {<NR1>|OFF|ON}
    - SAVEON:MEASUrement?
    - SAVEON:NUMEvents <NR3>
    - SAVEON:NUMEvents?
    - SAVEON:SETUP {<NR1>|ON|OFF}
    - SAVEON:SETUP?
    - SAVEON:TRIGger {<NR1>|ON|OFF}
    - SAVEON:TRIGger?
    - SAVEON:WAVEform {<NR1>|ON|OFF}
    - SAVEON:WAVEform?
    ```
"""

from typing import Optional, TYPE_CHECKING

from ..helpers import SCPICmdRead, SCPICmdWrite

if TYPE_CHECKING:
    from tm_devices.drivers.pi.pi_device import PIDevice


class SaveonWaveform(SCPICmdWrite, SCPICmdRead):
    """The ``SAVEON:WAVEform`` command.

    Description:
        - Sets or queries whether to save a waveform when a trigger occurs when ``SAVEON:TRIGGER``
          is ON. The waveform will be saved to the file you selected with ``SAVEON:FILE:NAME``, in
          the location that you selected using ``SAVEON:FILE:DEST``. You can set options for file
          storage (such as file name, file destination, and autoincrement), using the
          ``SAVEON:FILE`` commands.

    Usage:
        - Using the ``.query()`` method will send the ``SAVEON:WAVEform?`` query.
        - Using the ``.verify(value)`` method will send the ``SAVEON:WAVEform?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``SAVEON:WAVEform value`` command.

    SCPI Syntax:
        ```
        - SAVEON:WAVEform {<NR1>|ON|OFF}
        - SAVEON:WAVEform?
        ```

    Info:
        - ``<NR1>`` = 0 disables Save On Waveform; any other value turns this feature on.
        - ``OFF`` disables Save On Waveform.
        - ``ON`` enables Save On Waveform.
    """


class SaveonTrigger(SCPICmdWrite, SCPICmdRead):
    """The ``SAVEON:TRIGger`` command.

    Description:
        - Sets or queries whether to save a file when a trigger occurs. You can define the trigger
          using Trigger commands or the instrument user interface. This command is longer necessary.
          Please see Act On Event commands for future development. The trigger will cause the
          instrument to save an image or a waveform to a file, depending on what you specified. For
          example, if you have set ``SAVEON:IMAGE`` to On, and a trigger event occurs, the
          instrument will save a screen capture. You can set options for file storage (such as file
          name, file destination, and auto increment), using the ``SAVEON:FILE`` commands. Use the
          instrument interface to select whether to save one or more analog channels, digital
          channels, or math waveforms Analog and math waveforms are saved using one file per
          waveform. Digital waveforms are all saved to a single file.

    Usage:
        - Using the ``.query()`` method will send the ``SAVEON:TRIGger?`` query.
        - Using the ``.verify(value)`` method will send the ``SAVEON:TRIGger?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``SAVEON:TRIGger value`` command.

    SCPI Syntax:
        ```
        - SAVEON:TRIGger {<NR1>|ON|OFF}
        - SAVEON:TRIGger?
        ```

    Info:
        - ``<NR1>`` = 0 disables Save On Trigger; any other value turns this feature on.
        - ``OFF`` disables Save On Trigger.
        - ``ON`` enables Save On Trigger.
    """


class SaveonSetup(SCPICmdWrite, SCPICmdRead):
    """The ``SAVEON:SETUP`` command.

    Description:
        - Saves the instrument setup when there is an event. Queries the status of this feature (on
          or off).

    Usage:
        - Using the ``.query()`` method will send the ``SAVEON:SETUP?`` query.
        - Using the ``.verify(value)`` method will send the ``SAVEON:SETUP?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``SAVEON:SETUP value`` command.

    SCPI Syntax:
        ```
        - SAVEON:SETUP {<NR1>|ON|OFF}
        - SAVEON:SETUP?
        ```

    Info:
        - ``<NR1>`` = 0 disables the feature; any other value turns this feature on.
        - ``ON`` enables the feature.
        - ``OFF`` disables the feature.
    """


class SaveonNumevents(SCPICmdWrite, SCPICmdRead):
    """The ``SAVEON:NUMEvents`` command.

    Description:
        - Sets or queries the maximum number of events that will be saved. You can use this feature
          to avoid running out of disk space, especially if you save large files.

    Usage:
        - Using the ``.query()`` method will send the ``SAVEON:NUMEvents?`` query.
        - Using the ``.verify(value)`` method will send the ``SAVEON:NUMEvents?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``SAVEON:NUMEvents value`` command.

    SCPI Syntax:
        ```
        - SAVEON:NUMEvents <NR3>
        - SAVEON:NUMEvents?
        ```

    Info:
        - ``<NR3>`` specifies the number of events that will be saved before a reset is required.
    """


class SaveonMeasurement(SCPICmdWrite, SCPICmdRead):
    """The ``SAVEON:MEASUrement`` command.

    Description:
        - Sets or queries whether to save a measurement when any of the following triggers occurs:
          Limit test failure - if set to On. (``SAVEON:LIMIT``) Mask failure - if set to On.
          (``SAVEON:MASK``) Trigger - if set to On. (``SAVEON:TRIGGER``) The image will be saved to
          the file you selected with ``SAVEON:FILE:NAME``, in the location that you selected using
          ``SAVEON:FILE:DEST``. You can set options for file storage (such as file name, file
          destination, and autoincrement), using the SaveOn File commands.

    Usage:
        - Using the ``.query()`` method will send the ``SAVEON:MEASUrement?`` query.
        - Using the ``.verify(value)`` method will send the ``SAVEON:MEASUrement?`` query and raise
          an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``SAVEON:MEASUrement value`` command.

    SCPI Syntax:
        ```
        - SAVEON:MEASUrement {<NR1>|OFF|ON}
        - SAVEON:MEASUrement?
        ```

    Info:
        - ``<NR1>`` = 0 disables Save On Measurement; any other value turns this feature on.
        - ``OFF`` disables Save On Measurement.
        - ``ON`` enables Save On Measurement.
    """


class SaveonMask(SCPICmdWrite, SCPICmdRead):
    """The ``SAVEON:MASK`` command.

    Description:
        - Sets or queries whether to save a file when there is a mask failure. When ``SAVEON:MASK``
          is enabled, a mask failure will trigger the instrument to save an image, a measurement,
          and/or a waveform to a file, depending on what you specify using the Related Commands
          listed below. When there is a mask failure, if you have enabled ``SAVEON:IMAGe``, the
          instrument will save a screen capture to the file you selected using ``SAVEON:FILE:NAME``
          and ``SAVEON:FILE:DEST``. You can set options for file storage (such as file name, file
          destination, and autoincrement), using the SaveOn File commands. Use the oscilloscope
          interface to select whether to save one or more analog channels, digital channels, or math
          waveforms Analog and math waveforms are saved using one file per waveform. Digital
          waveforms are all saved to a single file.

    Usage:
        - Using the ``.query()`` method will send the ``SAVEON:MASK?`` query.
        - Using the ``.verify(value)`` method will send the ``SAVEON:MASK?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``SAVEON:MASK value`` command.

    SCPI Syntax:
        ```
        - SAVEON:MASK {<NR1>|OFF|ON}
        - SAVEON:MASK?
        ```

    Info:
        - ``<NR1>`` = 0 disables Save On Mask; any other value turns this feature on.
        - ``OFF`` disables Save On Mask.
        - ``ON`` enables Save On Mask.
    """


class SaveonLimit(SCPICmdWrite, SCPICmdRead):
    """The ``SAVEON:LIMit`` command.

    Description:
        - Sets or queries whether to save a file when there is a limit test failure. When this is
          set to On, a limit test failure will trigger the instrument to save an image, a
          measurement, and/or a waveform to a file, depending on what you specify using the Related
          Commands listed below. For example, if you set ``SAVEON:IMAGe`` to On, the instrument will
          save a screen capture to the file you selected with ``SAVEON:FILE:NAME``, in the location
          that you selected using ``SAVEON:FILE:DEST``. You can set options for file storage (such
          as file name, file destination, and autoincrement), using the SaveOn File commands. Analog
          and math waveforms are saved using one file per waveform. Digital waveforms are all saved
          to a single file.

    Usage:
        - Using the ``.query()`` method will send the ``SAVEON:LIMit?`` query.
        - Using the ``.verify(value)`` method will send the ``SAVEON:LIMit?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``SAVEON:LIMit value`` command.

    SCPI Syntax:
        ```
        - SAVEON:LIMit {<NR1>|OFF|ON}
        - SAVEON:LIMit?
        ```

    Info:
        - ``<NR1>`` = 0 disables Save On Limit; any other value turns this feature on.
        - ``OFF`` disables Save On Limit.
        - ``ON`` enables Save On Limit.
    """


class SaveonImage(SCPICmdWrite, SCPICmdRead):
    """The ``SAVEON:IMAGe`` command.

    Description:
        - This command sets or queries whether to save a screen capture when a trigger occurs and
          ``SAVEON:TRIGer`` is ON and ``SAVEON:IMAGE`` is ON.

    Usage:
        - Using the ``.query()`` method will send the ``SAVEON:IMAGe?`` query.
        - Using the ``.verify(value)`` method will send the ``SAVEON:IMAGe?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``SAVEON:IMAGe value`` command.

    SCPI Syntax:
        ```
        - SAVEON:IMAGe {<NR1>|OFF|ON}
        - SAVEON:IMAGe?
        ```

    Info:
        - ``<NR1>`` = 0 disables Save On Image; any other value turns this feature on.
        - ``OFF`` disables Save On Image.
        - ``ON`` enables Save On Image.
    """


class SaveonFileType(SCPICmdWrite, SCPICmdRead):
    """The ``SAVEON:FILE:TYPE`` command.

    Description:
        - Sets or queries whether to use the data and time as the file name (auto) or to use a
          custom file name.

    Usage:
        - Using the ``.query()`` method will send the ``SAVEON:FILE:TYPE?`` query.
        - Using the ``.verify(value)`` method will send the ``SAVEON:FILE:TYPE?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``SAVEON:FILE:TYPE value`` command.

    SCPI Syntax:
        ```
        - SAVEON:FILE:TYPE {AUTO|CUSTOM}
        - SAVEON:FILE:TYPE?
        ```

    Info:
        - ``AUTO`` uses the date and time as the file name for the saved events. An example auto
          file name is: ``20110711_182946``.
        - ``CUSTOM`` uses the file name that you specified using the ``SAVEON:FILE:NAME`` command.
    """


class SaveonFileName(SCPICmdWrite, SCPICmdRead):
    """The ``SAVEON:FILE:NAME`` command.

    Description:
        - Sets or queries the file name to use when the file type is set to Custom. Selecting a file
          name You can select your own file name by entering the desired name in the <string>.
          Follow standard Microsoft Windows naming conventions. The limit is 127 characters, with no
          spaces allowed in the name. File storage location The file is saved in the location
          specified by the ``SAVEON:FILE:DEST`` command. Default names The default file name is
          SaveOnEvent. This name will be appended under certain circumstances. If you have
          previously selected (through the oscilloscope interface) to save channel 1 and channel 2
          waveforms, for example, two files are saved, and the default file names will reflect the
          channel. If you have selected to save bus B1, the default file name will reflect that. The
          file name also indicates whether it is an image, measurement, or waveform. Autoincrement
          If ``SAVEON:FILE:AUTOInc`` is enabled, a number will be added to the file name and will be
          incremented for each new file. For example, if three events trigger the oscilloscope to
          save a measurement, three files will be created, which might be named 'SaveOnEvent1Meas,'
          'SaveOnEvent2Meas,' and 'SaveOnEvent3Meas.' Extensions Although these files have
          extensions (CSV for measurements and WFM for waveforms, for example), you do not need to
          enter the extension. It will be added automatically, depending on what you are saving. The
          query ``SAVEON:FILE:NAME?`` will return the file name without the extension.

    Usage:
        - Using the ``.query()`` method will send the ``SAVEON:FILE:NAME?`` query.
        - Using the ``.verify(value)`` method will send the ``SAVEON:FILE:NAME?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``SAVEON:FILE:NAME value`` command.

    SCPI Syntax:
        ```
        - SAVEON:FILE:NAME <string>
        - SAVEON:FILE:NAME?
        ```

    Info:
        - ``<string>`` is the file name you want to use.
    """


class SaveonFileDest(SCPICmdWrite, SCPICmdRead):
    """The ``SAVEON:FILE:DEST`` command.

    Description:
        - Sets or queries the location where files are saved. The default destination is
          C:usersusernameTektronixTekScopeSaveOnTrigger. You can save the files to a local drive or
          network path by entering the desired location in <string>. You can also select to save the
          files to a USB drive, using the browse button in the oscilloscope interface.

    Usage:
        - Using the ``.query()`` method will send the ``SAVEON:FILE:DEST?`` query.
        - Using the ``.verify(value)`` method will send the ``SAVEON:FILE:DEST?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``SAVEON:FILE:DEST value`` command.

    SCPI Syntax:
        ```
        - SAVEON:FILE:DEST <string>
        - SAVEON:FILE:DEST?
        ```
    """


class SaveonFileCount(SCPICmdWrite, SCPICmdRead):
    """The ``SAVEON:FILE:COUNt`` command.

    Description:
        - <primary></primary> Sets or queries the starting number that will be appended to the
          custom file name when ``SAVEON:FILE:AUTOINC`` is enabled and an image, measurement, or
          waveform is saved. The total of files saved cannot exceed 32767. If this file count is
          reached, files will not be saved until you change the file count to a lower number.

    Usage:
        - Using the ``.query()`` method will send the ``SAVEON:FILE:COUNt?`` query.
        - Using the ``.verify(value)`` method will send the ``SAVEON:FILE:COUNt?`` query and raise
          an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``SAVEON:FILE:COUNt value`` command.

    SCPI Syntax:
        ```
        - SAVEON:FILE:COUNt <NR3>
        - SAVEON:FILE:COUNt?
        ```

    Info:
        - ``<NR3>`` specifies the starting number for automatically incrementing the file name.
    """


class SaveonFileAutoinc(SCPICmdWrite, SCPICmdRead):
    """The ``SAVEON:FILE:AUTOInc`` command.

    Description:
        - Sets or queries the state of the auto increment file name feature (on or off). If AUTOInc
          is on, each time that an image, measurement, or waveform is saved to a file, the number
          that is appended to the file name will be augmented automatically, depending on the
          settings that you have made in the Related Commands, below. If AUTOInc is off, the
          previous file will be overwritten.

    Usage:
        - Using the ``.query()`` method will send the ``SAVEON:FILE:AUTOInc?`` query.
        - Using the ``.verify(value)`` method will send the ``SAVEON:FILE:AUTOInc?`` query and raise
          an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``SAVEON:FILE:AUTOInc value`` command.

    SCPI Syntax:
        ```
        - SAVEON:FILE:AUTOInc {<NR1>|ON|OFF}
        - SAVEON:FILE:AUTOInc?
        ```

    Info:
        - ``<NR1>`` = 1 enables the Auto increment function; any other value disables the function.
        - ``ON`` indicates that each time that an image, measurement, or waveform is saved, the
          number that is appended to the file name will be augmented automatically.
        - ``OFF`` indicates that when an image, measurement, or waveform file is saved, the name
          will not be augmented and the previously saved file will be overwritten.
    """


class SaveonFile(SCPICmdRead):
    """The ``SAVEON:FILE`` command tree.

    Usage:
        - Using the ``.query()`` method will send the ``SAVEON:FILE?`` query.
        - Using the ``.verify(value)`` method will send the ``SAVEON:FILE?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.autoinc``: The ``SAVEON:FILE:AUTOInc`` command.
        - ``.count``: The ``SAVEON:FILE:COUNt`` command.
        - ``.dest``: The ``SAVEON:FILE:DEST`` command.
        - ``.name``: The ``SAVEON:FILE:NAME`` command.
        - ``.type``: The ``SAVEON:FILE:TYPE`` command.
    """

    def __init__(self, device: Optional["PIDevice"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        self._autoinc = SaveonFileAutoinc(device, f"{self._cmd_syntax}:AUTOInc")
        self._count = SaveonFileCount(device, f"{self._cmd_syntax}:COUNt")
        self._dest = SaveonFileDest(device, f"{self._cmd_syntax}:DEST")
        self._name = SaveonFileName(device, f"{self._cmd_syntax}:NAME")
        self._type = SaveonFileType(device, f"{self._cmd_syntax}:TYPE")

    @property
    def autoinc(self) -> SaveonFileAutoinc:
        """Return the ``SAVEON:FILE:AUTOInc`` command.

        Description:
            - Sets or queries the state of the auto increment file name feature (on or off). If
              AUTOInc is on, each time that an image, measurement, or waveform is saved to a file,
              the number that is appended to the file name will be augmented automatically,
              depending on the settings that you have made in the Related Commands, below. If
              AUTOInc is off, the previous file will be overwritten.

        Usage:
            - Using the ``.query()`` method will send the ``SAVEON:FILE:AUTOInc?`` query.
            - Using the ``.verify(value)`` method will send the ``SAVEON:FILE:AUTOInc?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``SAVEON:FILE:AUTOInc value``
              command.

        SCPI Syntax:
            ```
            - SAVEON:FILE:AUTOInc {<NR1>|ON|OFF}
            - SAVEON:FILE:AUTOInc?
            ```

        Info:
            - ``<NR1>`` = 1 enables the Auto increment function; any other value disables the
              function.
            - ``ON`` indicates that each time that an image, measurement, or waveform is saved, the
              number that is appended to the file name will be augmented automatically.
            - ``OFF`` indicates that when an image, measurement, or waveform file is saved, the name
              will not be augmented and the previously saved file will be overwritten.
        """
        return self._autoinc

    @property
    def count(self) -> SaveonFileCount:
        """Return the ``SAVEON:FILE:COUNt`` command.

        Description:
            - <primary></primary> Sets or queries the starting number that will be appended to the
              custom file name when ``SAVEON:FILE:AUTOINC`` is enabled and an image, measurement, or
              waveform is saved. The total of files saved cannot exceed 32767. If this file count is
              reached, files will not be saved until you change the file count to a lower number.

        Usage:
            - Using the ``.query()`` method will send the ``SAVEON:FILE:COUNt?`` query.
            - Using the ``.verify(value)`` method will send the ``SAVEON:FILE:COUNt?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``SAVEON:FILE:COUNt value`` command.

        SCPI Syntax:
            ```
            - SAVEON:FILE:COUNt <NR3>
            - SAVEON:FILE:COUNt?
            ```

        Info:
            - ``<NR3>`` specifies the starting number for automatically incrementing the file name.
        """
        return self._count

    @property
    def dest(self) -> SaveonFileDest:
        """Return the ``SAVEON:FILE:DEST`` command.

        Description:
            - Sets or queries the location where files are saved. The default destination is
              C:usersusernameTektronixTekScopeSaveOnTrigger. You can save the files to a local drive
              or network path by entering the desired location in <string>. You can also select to
              save the files to a USB drive, using the browse button in the oscilloscope interface.

        Usage:
            - Using the ``.query()`` method will send the ``SAVEON:FILE:DEST?`` query.
            - Using the ``.verify(value)`` method will send the ``SAVEON:FILE:DEST?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``SAVEON:FILE:DEST value`` command.

        SCPI Syntax:
            ```
            - SAVEON:FILE:DEST <string>
            - SAVEON:FILE:DEST?
            ```
        """
        return self._dest

    @property
    def name(self) -> SaveonFileName:
        """Return the ``SAVEON:FILE:NAME`` command.

        Description:
            - Sets or queries the file name to use when the file type is set to Custom. Selecting a
              file name You can select your own file name by entering the desired name in the
              <string>. Follow standard Microsoft Windows naming conventions. The limit is 127
              characters, with no spaces allowed in the name. File storage location The file is
              saved in the location specified by the ``SAVEON:FILE:DEST`` command. Default names The
              default file name is SaveOnEvent. This name will be appended under certain
              circumstances. If you have previously selected (through the oscilloscope interface) to
              save channel 1 and channel 2 waveforms, for example, two files are saved, and the
              default file names will reflect the channel. If you have selected to save bus B1, the
              default file name will reflect that. The file name also indicates whether it is an
              image, measurement, or waveform. Autoincrement If ``SAVEON:FILE:AUTOInc`` is enabled,
              a number will be added to the file name and will be incremented for each new file. For
              example, if three events trigger the oscilloscope to save a measurement, three files
              will be created, which might be named 'SaveOnEvent1Meas,' 'SaveOnEvent2Meas,' and
              'SaveOnEvent3Meas.' Extensions Although these files have extensions (CSV for
              measurements and WFM for waveforms, for example), you do not need to enter the
              extension. It will be added automatically, depending on what you are saving. The query
              ``SAVEON:FILE:NAME?`` will return the file name without the extension.

        Usage:
            - Using the ``.query()`` method will send the ``SAVEON:FILE:NAME?`` query.
            - Using the ``.verify(value)`` method will send the ``SAVEON:FILE:NAME?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``SAVEON:FILE:NAME value`` command.

        SCPI Syntax:
            ```
            - SAVEON:FILE:NAME <string>
            - SAVEON:FILE:NAME?
            ```

        Info:
            - ``<string>`` is the file name you want to use.
        """
        return self._name

    @property
    def type(self) -> SaveonFileType:
        """Return the ``SAVEON:FILE:TYPE`` command.

        Description:
            - Sets or queries whether to use the data and time as the file name (auto) or to use a
              custom file name.

        Usage:
            - Using the ``.query()`` method will send the ``SAVEON:FILE:TYPE?`` query.
            - Using the ``.verify(value)`` method will send the ``SAVEON:FILE:TYPE?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``SAVEON:FILE:TYPE value`` command.

        SCPI Syntax:
            ```
            - SAVEON:FILE:TYPE {AUTO|CUSTOM}
            - SAVEON:FILE:TYPE?
            ```

        Info:
            - ``AUTO`` uses the date and time as the file name for the saved events. An example auto
              file name is: ``20110711_182946``.
            - ``CUSTOM`` uses the file name that you specified using the ``SAVEON:FILE:NAME``
              command.
        """
        return self._type


class SaveonCount(SCPICmdRead):
    """The ``SAVEON:COUNt`` command.

    Description:
        - Returns the number of events (files) that have been saved since the last reset.

    Usage:
        - Using the ``.query()`` method will send the ``SAVEON:COUNt?`` query.
        - Using the ``.verify(value)`` method will send the ``SAVEON:COUNt?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    SCPI Syntax:
        ```
        - SAVEON:COUNt?
        ```
    """


#  pylint: disable=too-many-instance-attributes
class Saveon(SCPICmdWrite, SCPICmdRead):
    """The ``SAVEON`` command.

    Description:
        - Sets the auto-increment file count to 0. Once the number of saved files has reached the
          limit that you set (using the ``SAVEON:NUMevents`` command), no files will be saved until
          you reset the count.

    Usage:
        - Using the ``.write(value)`` method will send the ``SAVEON value`` command.

    SCPI Syntax:
        ```
        - SAVEON {RESET}
        ```

    Info:
        - ``RESET`` sets the file count to 0.

    Properties:
        - ``.count``: The ``SAVEON:COUNt`` command.
        - ``.file``: The ``SAVEON:FILE`` command tree.
        - ``.image``: The ``SAVEON:IMAGe`` command.
        - ``.limit``: The ``SAVEON:LIMit`` command.
        - ``.mask``: The ``SAVEON:MASK`` command.
        - ``.measurement``: The ``SAVEON:MEASUrement`` command.
        - ``.numevents``: The ``SAVEON:NUMEvents`` command.
        - ``.setup``: The ``SAVEON:SETUP`` command.
        - ``.trigger``: The ``SAVEON:TRIGger`` command.
        - ``.waveform``: The ``SAVEON:WAVEform`` command.
    """

    def __init__(self, device: Optional["PIDevice"] = None, cmd_syntax: str = "SAVEON") -> None:
        super().__init__(device, cmd_syntax)
        self._count = SaveonCount(device, f"{self._cmd_syntax}:COUNt")
        self._file = SaveonFile(device, f"{self._cmd_syntax}:FILE")
        self._image = SaveonImage(device, f"{self._cmd_syntax}:IMAGe")
        self._limit = SaveonLimit(device, f"{self._cmd_syntax}:LIMit")
        self._mask = SaveonMask(device, f"{self._cmd_syntax}:MASK")
        self._measurement = SaveonMeasurement(device, f"{self._cmd_syntax}:MEASUrement")
        self._numevents = SaveonNumevents(device, f"{self._cmd_syntax}:NUMEvents")
        self._setup = SaveonSetup(device, f"{self._cmd_syntax}:SETUP")
        self._trigger = SaveonTrigger(device, f"{self._cmd_syntax}:TRIGger")
        self._waveform = SaveonWaveform(device, f"{self._cmd_syntax}:WAVEform")

    @property
    def count(self) -> SaveonCount:
        """Return the ``SAVEON:COUNt`` command.

        Description:
            - Returns the number of events (files) that have been saved since the last reset.

        Usage:
            - Using the ``.query()`` method will send the ``SAVEON:COUNt?`` query.
            - Using the ``.verify(value)`` method will send the ``SAVEON:COUNt?`` query and raise an
              AssertionError if the returned value does not match ``value``.

        SCPI Syntax:
            ```
            - SAVEON:COUNt?
            ```
        """
        return self._count

    @property
    def file(self) -> SaveonFile:
        """Return the ``SAVEON:FILE`` command tree.

        Usage:
            - Using the ``.query()`` method will send the ``SAVEON:FILE?`` query.
            - Using the ``.verify(value)`` method will send the ``SAVEON:FILE?`` query and raise an
              AssertionError if the returned value does not match ``value``.

        Sub-properties:
            - ``.autoinc``: The ``SAVEON:FILE:AUTOInc`` command.
            - ``.count``: The ``SAVEON:FILE:COUNt`` command.
            - ``.dest``: The ``SAVEON:FILE:DEST`` command.
            - ``.name``: The ``SAVEON:FILE:NAME`` command.
            - ``.type``: The ``SAVEON:FILE:TYPE`` command.
        """
        return self._file

    @property
    def image(self) -> SaveonImage:
        """Return the ``SAVEON:IMAGe`` command.

        Description:
            - This command sets or queries whether to save a screen capture when a trigger occurs
              and ``SAVEON:TRIGer`` is ON and ``SAVEON:IMAGE`` is ON.

        Usage:
            - Using the ``.query()`` method will send the ``SAVEON:IMAGe?`` query.
            - Using the ``.verify(value)`` method will send the ``SAVEON:IMAGe?`` query and raise an
              AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``SAVEON:IMAGe value`` command.

        SCPI Syntax:
            ```
            - SAVEON:IMAGe {<NR1>|OFF|ON}
            - SAVEON:IMAGe?
            ```

        Info:
            - ``<NR1>`` = 0 disables Save On Image; any other value turns this feature on.
            - ``OFF`` disables Save On Image.
            - ``ON`` enables Save On Image.
        """
        return self._image

    @property
    def limit(self) -> SaveonLimit:
        """Return the ``SAVEON:LIMit`` command.

        Description:
            - Sets or queries whether to save a file when there is a limit test failure. When this
              is set to On, a limit test failure will trigger the instrument to save an image, a
              measurement, and/or a waveform to a file, depending on what you specify using the
              Related Commands listed below. For example, if you set ``SAVEON:IMAGe`` to On, the
              instrument will save a screen capture to the file you selected with
              ``SAVEON:FILE:NAME``, in the location that you selected using ``SAVEON:FILE:DEST``.
              You can set options for file storage (such as file name, file destination, and
              autoincrement), using the SaveOn File commands. Analog and math waveforms are saved
              using one file per waveform. Digital waveforms are all saved to a single file.

        Usage:
            - Using the ``.query()`` method will send the ``SAVEON:LIMit?`` query.
            - Using the ``.verify(value)`` method will send the ``SAVEON:LIMit?`` query and raise an
              AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``SAVEON:LIMit value`` command.

        SCPI Syntax:
            ```
            - SAVEON:LIMit {<NR1>|OFF|ON}
            - SAVEON:LIMit?
            ```

        Info:
            - ``<NR1>`` = 0 disables Save On Limit; any other value turns this feature on.
            - ``OFF`` disables Save On Limit.
            - ``ON`` enables Save On Limit.
        """
        return self._limit

    @property
    def mask(self) -> SaveonMask:
        """Return the ``SAVEON:MASK`` command.

        Description:
            - Sets or queries whether to save a file when there is a mask failure. When
              ``SAVEON:MASK`` is enabled, a mask failure will trigger the instrument to save an
              image, a measurement, and/or a waveform to a file, depending on what you specify using
              the Related Commands listed below. When there is a mask failure, if you have enabled
              ``SAVEON:IMAGe``, the instrument will save a screen capture to the file you selected
              using ``SAVEON:FILE:NAME`` and ``SAVEON:FILE:DEST``. You can set options for file
              storage (such as file name, file destination, and autoincrement), using the SaveOn
              File commands. Use the oscilloscope interface to select whether to save one or more
              analog channels, digital channels, or math waveforms Analog and math waveforms are
              saved using one file per waveform. Digital waveforms are all saved to a single file.

        Usage:
            - Using the ``.query()`` method will send the ``SAVEON:MASK?`` query.
            - Using the ``.verify(value)`` method will send the ``SAVEON:MASK?`` query and raise an
              AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``SAVEON:MASK value`` command.

        SCPI Syntax:
            ```
            - SAVEON:MASK {<NR1>|OFF|ON}
            - SAVEON:MASK?
            ```

        Info:
            - ``<NR1>`` = 0 disables Save On Mask; any other value turns this feature on.
            - ``OFF`` disables Save On Mask.
            - ``ON`` enables Save On Mask.
        """
        return self._mask

    @property
    def measurement(self) -> SaveonMeasurement:
        """Return the ``SAVEON:MEASUrement`` command.

        Description:
            - Sets or queries whether to save a measurement when any of the following triggers
              occurs: Limit test failure - if set to On. (``SAVEON:LIMIT``) Mask failure - if set to
              On. (``SAVEON:MASK``) Trigger - if set to On. (``SAVEON:TRIGGER``) The image will be
              saved to the file you selected with ``SAVEON:FILE:NAME``, in the location that you
              selected using ``SAVEON:FILE:DEST``. You can set options for file storage (such as
              file name, file destination, and autoincrement), using the SaveOn File commands.

        Usage:
            - Using the ``.query()`` method will send the ``SAVEON:MEASUrement?`` query.
            - Using the ``.verify(value)`` method will send the ``SAVEON:MEASUrement?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``SAVEON:MEASUrement value`` command.

        SCPI Syntax:
            ```
            - SAVEON:MEASUrement {<NR1>|OFF|ON}
            - SAVEON:MEASUrement?
            ```

        Info:
            - ``<NR1>`` = 0 disables Save On Measurement; any other value turns this feature on.
            - ``OFF`` disables Save On Measurement.
            - ``ON`` enables Save On Measurement.
        """
        return self._measurement

    @property
    def numevents(self) -> SaveonNumevents:
        """Return the ``SAVEON:NUMEvents`` command.

        Description:
            - Sets or queries the maximum number of events that will be saved. You can use this
              feature to avoid running out of disk space, especially if you save large files.

        Usage:
            - Using the ``.query()`` method will send the ``SAVEON:NUMEvents?`` query.
            - Using the ``.verify(value)`` method will send the ``SAVEON:NUMEvents?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``SAVEON:NUMEvents value`` command.

        SCPI Syntax:
            ```
            - SAVEON:NUMEvents <NR3>
            - SAVEON:NUMEvents?
            ```

        Info:
            - ``<NR3>`` specifies the number of events that will be saved before a reset is
              required.
        """
        return self._numevents

    @property
    def setup(self) -> SaveonSetup:
        """Return the ``SAVEON:SETUP`` command.

        Description:
            - Saves the instrument setup when there is an event. Queries the status of this feature
              (on or off).

        Usage:
            - Using the ``.query()`` method will send the ``SAVEON:SETUP?`` query.
            - Using the ``.verify(value)`` method will send the ``SAVEON:SETUP?`` query and raise an
              AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``SAVEON:SETUP value`` command.

        SCPI Syntax:
            ```
            - SAVEON:SETUP {<NR1>|ON|OFF}
            - SAVEON:SETUP?
            ```

        Info:
            - ``<NR1>`` = 0 disables the feature; any other value turns this feature on.
            - ``ON`` enables the feature.
            - ``OFF`` disables the feature.
        """
        return self._setup

    @property
    def trigger(self) -> SaveonTrigger:
        """Return the ``SAVEON:TRIGger`` command.

        Description:
            - Sets or queries whether to save a file when a trigger occurs. You can define the
              trigger using Trigger commands or the instrument user interface. This command is
              longer necessary. Please see Act On Event commands for future development. The trigger
              will cause the instrument to save an image or a waveform to a file, depending on what
              you specified. For example, if you have set ``SAVEON:IMAGE`` to On, and a trigger
              event occurs, the instrument will save a screen capture. You can set options for file
              storage (such as file name, file destination, and auto increment), using the
              ``SAVEON:FILE`` commands. Use the instrument interface to select whether to save one
              or more analog channels, digital channels, or math waveforms Analog and math waveforms
              are saved using one file per waveform. Digital waveforms are all saved to a single
              file.

        Usage:
            - Using the ``.query()`` method will send the ``SAVEON:TRIGger?`` query.
            - Using the ``.verify(value)`` method will send the ``SAVEON:TRIGger?`` query and raise
              an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``SAVEON:TRIGger value`` command.

        SCPI Syntax:
            ```
            - SAVEON:TRIGger {<NR1>|ON|OFF}
            - SAVEON:TRIGger?
            ```

        Info:
            - ``<NR1>`` = 0 disables Save On Trigger; any other value turns this feature on.
            - ``OFF`` disables Save On Trigger.
            - ``ON`` enables Save On Trigger.
        """
        return self._trigger

    @property
    def waveform(self) -> SaveonWaveform:
        """Return the ``SAVEON:WAVEform`` command.

        Description:
            - Sets or queries whether to save a waveform when a trigger occurs when
              ``SAVEON:TRIGGER`` is ON. The waveform will be saved to the file you selected with
              ``SAVEON:FILE:NAME``, in the location that you selected using ``SAVEON:FILE:DEST``.
              You can set options for file storage (such as file name, file destination, and
              autoincrement), using the ``SAVEON:FILE`` commands.

        Usage:
            - Using the ``.query()`` method will send the ``SAVEON:WAVEform?`` query.
            - Using the ``.verify(value)`` method will send the ``SAVEON:WAVEform?`` query and raise
              an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``SAVEON:WAVEform value`` command.

        SCPI Syntax:
            ```
            - SAVEON:WAVEform {<NR1>|ON|OFF}
            - SAVEON:WAVEform?
            ```

        Info:
            - ``<NR1>`` = 0 disables Save On Waveform; any other value turns this feature on.
            - ``OFF`` disables Save On Waveform.
            - ``ON`` enables Save On Waveform.
        """
        return self._waveform
